/*
 * Copyright 2006-2007, Thomas Maier-Komor
 * LICENSE: GPLv2, see file COPYING for details
 */

#include <alloca.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <kstat.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <sys/loadavg.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/swap.h>
#include <sys/systeminfo.h>
#include <sys/sysinfo.h>
#include <time.h>
#include <unistd.h>
#include <utmpx.h>


/* SUSv3 does not have PATH_MAX */
#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

#include "log.h"
#include "network.h"
#include "sysstat.h"

extern int Localmode;
static kstat_ctl_t *KS = 0;
static sysstat_t Sysstat;


typedef struct cpudata_s {
	long long 
		user, kernel, idle, wait,
		intr, intrthread, intrblk, xcalls,
		syscall, sysexec, sysfork, sysvfork,
		sysread, syswrite,
		pgswapin, pgswapout, scan,
		anonpgin, anonpgout, anonfree,
		as_fault, hat_fault, majfault,
		phread, phwrite,
		rwrf, rwwf, mxae,
		inv_swtch, pswitch, migrate;
	hrtime_t
		user_ts, kernel_ts, idle_ts, wait_ts,
		timestamp;
	kstat_t 
		*cpustat, *cpusys;
	cpurates_t
		*rates;
	unsigned
		sys_iter, stat_iter;
} cpudata_t;


typedef struct class_chain_s {
	kstat_t *kstat;
	unsigned long long value[2];
	hrtime_t timestamp;
	unsigned iteration;
	struct class_chain_s *next;
} class_chain_t;

/* the number of cpus can change dynamically under Solaris */
static unsigned long
	Ocpu = 0,	/* number of online cpus */
	Maxid = 0,	/* highest id of a cpu */
	Iteration = 1,	/* used to make sure data is not update twice after a reconfig */
	Ncpus = 0;	/* number of CPUs available in the system (not necessaryly online) */

static cpudata_t
	**Cpus = 0;
static cpurates_t
	*Cpurates = 0;
kstat_t
	*Syspages = 0, *Ipstat = 0;

static class_chain_t
	*Err = 0, *Disks = 0, *Nfs = 0, *NfsServer = 0, *Net = 0;

/* cpu/sys percentages */
double UserP, KernelP, IdleP;
/* cpu/sys rates */
double IntrR, IntrThreadR, IntrBlkR, XcallsR, SyscallR, SysexecR, SysforkR,
       SysreadR, SyswriteR, LockfR;
/* cpu/vm rates */
double InvSwtchR, PswitchR, MigrateR, AsFaultR, HatFaultR, MajFaultR,
       PgswapinR, PgswapoutR, ScanR, AnonfreeR, AnonpginR, AnonpgoutR;
/* ip */
double IpInR, IpOutR, NetInR, NetOutR;
unsigned long IpInDiscards = 0, IpOutDiscards = 0;
/* unix/system_pages */
unsigned long Physmem, Pagesfree, Pageslocked, Pagestotal;
unsigned long Pgsz;
/* swap (non-kstat info) */
unsigned long SwapTotal, SwapFree;
/* device_errors */
unsigned long long Devices = 0, Storage;
unsigned long HardErrors = 0, SoftErrors = 0;
/* device I/O rates */
double DiskIR, DiskOR;
/* nfs I/O rates */
double NfsIR, NfsOR;
/* nfsd op rate */
double NfsdR;
/* boot time */
clock_t BootTime;

/* ip */
unsigned long long IpInReceives, IpOutRequests;
unsigned long IpInErrs;
hrtime_t IpIn_ts, IpOut_ts;



static void
get_boottime(void)
{
	struct utmpx utmpx, *ent;
	utmpx.ut_type = BOOT_TIME;
	ent = getutxid(&utmpx);
	assert(ent);
	BootTime = ent->ut_tv.tv_sec;
}


static void
get_swapinfo(void)
{
	swaptbl_t *swaptbl;
	char *strtbl;
	int numswap,i;

	numswap = swapctl(SC_GETNSWP,0);
	assert(numswap != -1);
	numswap += 2;
	swaptbl = alloca(numswap * sizeof(swapent_t) + sizeof(struct swaptable));
	swaptbl->swt_n = numswap;
	strtbl = alloca(numswap * PATH_MAX);
	for (i = 0; i < numswap; ++i)
		swaptbl->swt_ent[i].ste_path = strtbl + (i * PATH_MAX);
	numswap = swapctl(SC_LIST,swaptbl);
	/* If you encounter the following assertion, then please
	 * recompile in 64 bit mode. It is caused by an interface
	 * limitation.
	 */
	assert(numswap != -1);
	SwapTotal = SwapFree = 0;
	for (i = 0; i < numswap; ++i) {
		SwapTotal += swaptbl->swt_ent[i].ste_pages;
		SwapFree += swaptbl->swt_ent[i].ste_free;
	}
	SwapTotal = ((long long)SwapTotal * Pgsz) >> 10;
	SwapFree = ((long long)SwapFree * Pgsz) >> 10;
}


static class_chain_t *
get_class_chain(const char *class, const char *module, const char *name, class_chain_t *c)
{
	kstat_t *k;
	class_chain_t *p = 0, *f = c;
	if (0 == c) {
		c = calloc(1,sizeof(class_chain_t));
		assert(c);
		f = c;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if ((class && strcmp(class,k->ks_class)) ||
		    (module && strcmp(module,k->ks_module)) ||
		    (name && strncmp(name,k->ks_name,strlen(name))))
			continue;
		if (0 == c) {
			c = calloc(1,sizeof(class_chain_t));
			/* p is always a valid pointer, because
			 * we allocate c before going into this loop
			 * and in every iteration if necessary
			 */
			p->next = c;
		} else {
			c->value[0] = c->value[1] = 0;
		}
		dbug("get_class_chain(): adding %s/%s/%s\n",k->ks_class,k->ks_module,k->ks_name);
		c->kstat = k;
		p = c;
		c = c->next;
	}
	while (c) {
		c->kstat = 0;
		c = c->next;
	}
	return f;
}


/* Solaris >=10 only */
static void
get_ip_ip(void)
{
	kstat_named_t *data;
	kid_t kid;

	kid = kstat_read(KS,Ipstat,0);
	assert(kid != -1);

	data = kstat_data_lookup(Ipstat,"inReceives");
	assert(data);
	if (data->data_type == KSTAT_DATA_UINT32) {
		IpInR = (double)(data->value.ui32 - IpInReceives) * 1E9 /
			(double)(Ipstat->ks_snaptime - IpIn_ts);
		IpIn_ts = Ipstat->ks_snaptime;
		IpInReceives = data->value.ui32;
	} else if (data->data_type == KSTAT_DATA_UINT64) {
		IpInR = (double)(data->value.ui64 - IpInReceives) * 1E9 /
			(double)(Ipstat->ks_snaptime - IpIn_ts);
		IpIn_ts = Ipstat->ks_snaptime;
		IpInReceives = data->value.ui64;
	} else
		assert(0);

	data = kstat_data_lookup(Ipstat,"outRequests");
	assert(data);

	if (data->data_type == KSTAT_DATA_UINT32) {
		IpOutR = (double)(data->value.ui32 - IpOutRequests) * 1E9/
			(double)(Ipstat->ks_snaptime - IpOut_ts);
		IpOut_ts = Ipstat->ks_snaptime;
		IpOutRequests = data->value.ui32;
	} else if (data->data_type == KSTAT_DATA_UINT64) {
		IpOutR = (double)(data->value.ui64 - IpOutRequests) * 1E9/
			(double)(Ipstat->ks_snaptime - IpOut_ts);
		IpOut_ts = Ipstat->ks_snaptime;
		IpOutRequests = data->value.ui64;
	} else 
		assert(0);

	data = kstat_data_lookup(Ipstat,"inDiscards");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	IpInDiscards = data->value.ui32;

	data = kstat_data_lookup(Ipstat,"outDiscards");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	IpOutDiscards = data->value.ui32;
}


static int
get_device_errors(void)
{
	class_chain_t *c;

	HardErrors = 0;
	SoftErrors = 0;
	Storage = 0;
	Devices = 0;
	for (c = Err; c; c = c->next) {
		kstat_named_t *data;

		if (-1 == kstat_read(KS,c->kstat,0))
			return 1;
		assert(c->kstat->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(c->kstat,"Size");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		if (data->value.ui64 == 0)
			continue;
		++Devices;
		Storage += data->value.ui64;
		data = kstat_data_lookup(c->kstat,"Hard Errors");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT32);
		HardErrors += data->value.ui32;
		data = kstat_data_lookup(c->kstat,"Soft Errors");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT32);
		SoftErrors += data->value.ui32;
	}
	Storage >>= 10;
	return 0;
}


static int
get_disk_io(void)
{
	class_chain_t *c;
	Devices = 0;
	for (c = Disks; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_io_t *data;
		double td;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert((s->ks_type == KSTAT_TYPE_IO)
			&& (s->ks_snaptime != c->timestamp));
		data = s->ks_data;
		td = 1E9 / (s->ks_snaptime - c->timestamp);
		if (c->value[0])
			DiskIR += (double)(data->nread - c->value[0]) * td;
		if (c->value[1])
			DiskOR += (double)(data->nwritten - c->value[1]) * td;
		c->value[0] = data->nread;
		c->value[1] = data->nwritten;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	return 0;
}


static int
get_nfs_io(void)
{
	class_chain_t *c;
	for (c = Nfs; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_io_t *data;
		double td;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert(s->ks_type == KSTAT_TYPE_IO);
		data = s->ks_data;
		assert(s->ks_snaptime != c->timestamp);
		td = 1E9 / (s->ks_snaptime - c->timestamp);
		if (c->value[0])
			NfsIR += (double)(data->nread - c->value[0]) * td;
		assert(NfsIR < 1E10);
		if (c->value[1])
			NfsOR += (double)(data->nwritten - c->value[1]) * td;
		assert(NfsOR < 1E10);
		c->value[0] = data->nread;
		c->value[1] = data->nwritten;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	for (c = NfsServer; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_named_t *data;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert(s->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(s,"calls");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (c->value[0])
			NfsdR += (double)(data->value.ui64 - c->value[0]) * 1E9 /
				(s->ks_snaptime - c->timestamp);
		c->value[0] = data->value.ui64;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	return 0;
}


static int
get_unix_systempages(void)
{
	kstat_named_t *data;

	if (kstat_read(KS,Syspages,0) == -1) {
		/* cannot be unconfigured/cannot fail */
		assert(0);
	}

	data = kstat_data_lookup(Syspages,"physmem");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Physmem = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Physmem = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pagesfree");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Pagesfree = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Pagesfree = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pageslocked");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Pageslocked = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Pageslocked = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pagestotal");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Pagestotal = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Pagestotal = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);
	return 0;
}


/* this is only for Solaris 10 and later */
static int
get_cpu_sys(void)
{
	int inst = 0, num = 0;
	cpurates_t *nextrates = Cpurates;

	/* not here, because for Solaris 9 this must be done in get_cpu_stat
	 * IdleP = UserP = KernelP = 0;
	 */
	for (inst = 0; inst < Maxid; ++inst) {
		kstat_named_t *data;
		cpudata_t *cpu;
		cpurates_t *rates;
		kstat_t *cpusys;
		unsigned long long user, kernel, idle;
		hrtime_t user_ts, kernel_ts, idle_ts;

		cpu = Cpus[inst];
		rates = cpu->rates;
		cpusys = cpu->cpusys;
		if (cpusys == 0)
			continue;
		if (kstat_read(KS,cpusys,0) == -1) {
			/* processor has been taken offline */
			if (rates) {
				free(rates);
				cpu->rates = 0;
			}
			Cpus[inst]->cpusys = 0;
			return 1;
		}
		++num;
		if (cpu->sys_iter == Iteration) {
			/* fields already up-to-date */
			continue;
		}
		cpu->sys_iter = Iteration;
		if (rates == 0)
			return 1;
		rates->next = nextrates;
		nextrates = rates;

		/* Every lookup here may fail, but the whole set of metrics
		 * should be committed as a transaction, because otherwise
		 * we won't be able to handle dynamic reconfiguration
		 * correctly. So first do all lookups and then write the
		 * results back. Every lookup modifies the ks_snaptime, so
		 * we need to have seperate timestamps.
		 */
		data = kstat_data_lookup(cpusys,"cpu_nsec_user");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		user = data->value.ui64;
		user_ts = cpusys->ks_snaptime;
		data = kstat_data_lookup(cpusys,"cpu_nsec_kernel");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		kernel = data->value.ui64;
		kernel_ts = cpusys->ks_snaptime;
		data = kstat_data_lookup(cpusys,"cpu_nsec_idle");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		idle = data->value.ui64;
		idle_ts = cpusys->ks_snaptime;

		rates->user = (double)(user - cpu->user)
			/ (double)(user_ts - cpu->user_ts)
			* 100;
		UserP += rates->user / Ocpu;
		cpu->user = user;
		cpu->user_ts = user_ts;
		rates->kernel = (double)(kernel - cpu->kernel)
			/ (double)(kernel_ts - cpu->kernel_ts)
			* 100;
		KernelP += rates->kernel / Ocpu;
		cpu->kernel = kernel;
		cpu->kernel_ts = kernel_ts;
		rates->idle = (double)(idle - cpu->idle)
			/ (double)(idle_ts - cpu->idle_ts)
			* 100;
		IdleP += rates->idle / Ocpu;
		cpu->idle = idle;
		cpu->idle_ts = idle_ts;
	}
	Cpurates = nextrates;
	if (num >= Ocpu)
		return 0;
	return 1;
}


static int
get_cpu_stat(void)
{
	unsigned inst = 0, num = 0;

	for (inst = 0; inst < Maxid; ++inst) {
		cpudata_t *cpu = Cpus[inst];
		kstat_t *cpustat = cpu->cpustat;
		cpurates_t *rates = cpu->rates;
		cpu_stat_t *stat;
		double td;

		if (cpustat == 0)
			continue;
		if (-1 == kstat_read(KS,cpustat,0)) {
			/* processor was taken offline */
			return 1;
		}

		++num;
		if (cpu->stat_iter == Iteration) {
			/* this info was updated before reconfiguring */
			continue;
		}
		stat = cpustat->ks_data;
		cpu->stat_iter = Iteration;
		td = 1 / ((cpustat->ks_snaptime - cpu->timestamp) * 1E-9);

		if (rates == 0) {
			rates = calloc(1,sizeof(cpurates_t));
			cpu->rates = rates;
		}
		rates->intr = (double)(stat->cpu_sysinfo.intr - cpu->intr) * td;
		IntrR += rates->intr;
		cpu->intr = stat->cpu_sysinfo.intr;
		rates->intrthread = (double)(stat->cpu_sysinfo.intrthread - cpu->intrthread) * td;
		IntrThreadR += rates->intrthread;
		cpu->intrthread = stat->cpu_sysinfo.intrthread;
		rates->intrblk = (double)(stat->cpu_sysinfo.intrblk - cpu->intrblk) * td;
		IntrBlkR += rates->intrblk;
		cpu->intrblk = stat->cpu_sysinfo.intrblk;
		rates->xcalls = (double)(stat->cpu_sysinfo.xcalls - cpu->xcalls) * td;
		XcallsR += rates->xcalls;
		cpu->xcalls = stat->cpu_sysinfo.xcalls;
		rates->pswitch = (double)(stat->cpu_sysinfo.pswitch - cpu->pswitch) * td;
		PswitchR += rates->pswitch;
		cpu->pswitch = stat->cpu_sysinfo.pswitch;
		rates->migrate = (double)(stat->cpu_sysinfo.cpumigrate - cpu->migrate) * td;
		MigrateR += rates->migrate;
		cpu->migrate = stat->cpu_sysinfo.cpumigrate;
		rates->inv_swtch = (double)(stat->cpu_sysinfo.inv_swtch - cpu->inv_swtch) * td;
		InvSwtchR += rates->inv_swtch;
		cpu->inv_swtch = stat->cpu_sysinfo.inv_swtch;
		rates->syscall = (double)(stat->cpu_sysinfo.syscall - cpu->syscall) * td;
		SyscallR += rates->syscall;
		cpu->syscall = stat->cpu_sysinfo.syscall;
		rates->sysfork = (double)(stat->cpu_sysinfo.sysfork - cpu->sysfork) * td;
		cpu->sysfork = stat->cpu_sysinfo.sysfork;
		rates->sysfork += (double)(stat->cpu_sysinfo.sysvfork - cpu->sysvfork) * td;
		SysforkR += rates->sysfork;
		cpu->sysvfork = stat->cpu_sysinfo.sysvfork;
		rates->sysexec = (double)(stat->cpu_sysinfo.sysexec - cpu->sysexec) * td;
		SysexecR += rates->sysexec;
		cpu->sysexec = stat->cpu_sysinfo.sysexec;
		rates->sysread = (double)(stat->cpu_sysinfo.sysread - cpu->sysread) * td;
		SysreadR += rates->sysread;
		cpu->sysread = stat->cpu_sysinfo.sysread;
		rates->syswrite = (double)(stat->cpu_sysinfo.syswrite - cpu->syswrite) * td;
		SyswriteR += rates->syswrite;
		cpu->syswrite = stat->cpu_sysinfo.syswrite;
		rates->anonpgin = (double)(stat->cpu_vminfo.anonpgin - cpu->anonpgin) * td;
		AnonpginR += rates->anonpgin;
		cpu->anonpgin = stat->cpu_vminfo.anonpgin;
		rates->anonpgout = (double)(stat->cpu_vminfo.anonpgout - cpu->anonpgout) * td;
		AnonpgoutR += rates->anonpgout;
		cpu->anonpgout = stat->cpu_vminfo.anonpgout;
		rates->pgswapin = (double)(stat->cpu_vminfo.pgswapin - cpu->pgswapin) * td;
		PgswapinR += rates->pgswapin;
		cpu->pgswapin = stat->cpu_vminfo.pgswapin;
		rates->pgswapout = (double)(stat->cpu_vminfo.pgswapout - cpu->pgswapout) * td;
		PgswapoutR += rates->pgswapout;
		cpu->pgswapout = stat->cpu_vminfo.pgswapout;
		rates->scan = (double)(stat->cpu_vminfo.scan - cpu->scan) * td;
		ScanR += rates->scan;
		cpu->scan = stat->cpu_vminfo.scan;
		rates->anonfree = (double)(stat->cpu_vminfo.anonfree - cpu->anonfree) * td;
		AnonfreeR += rates->anonfree;
		cpu->anonfree = stat->cpu_vminfo.anonfree;
		rates->as_fault = (double)(stat->cpu_vminfo.as_fault - cpu->as_fault) * td;
		AsFaultR += rates->as_fault;
		cpu->as_fault = stat->cpu_vminfo.as_fault;
		rates->hat_fault = (double)(stat->cpu_vminfo.hat_fault - cpu->hat_fault) * td;
		HatFaultR += rates->hat_fault;
		cpu->hat_fault = stat->cpu_vminfo.hat_fault;
		rates->majfault = (double)(stat->cpu_vminfo.maj_fault - cpu->majfault) * td;
		MajFaultR += rates->majfault;
		cpu->majfault = stat->cpu_vminfo.maj_fault;
		rates->rwrf = (double)(stat->cpu_sysinfo.rw_rdfails - cpu->rwrf) * td;
		LockfR += rates->rwrf;
		cpu->rwrf = stat->cpu_sysinfo.rw_rdfails;
		rates->rwwf = (double)(stat->cpu_sysinfo.rw_wrfails - cpu->rwwf) * td;
		LockfR += rates->rwwf;
		cpu->rwwf = stat->cpu_sysinfo.rw_wrfails;
		rates->mxae = (double)(stat->cpu_sysinfo.mutex_adenters - cpu->mxae) * td;
		LockfR += rates->mxae;
		cpu->mxae = stat->cpu_sysinfo.mutex_adenters;
		cpu->timestamp = cpustat->ks_snaptime;

		if (Cpus[inst]->cpusys == 0) {
			int total = stat->cpu_sysinfo.cpu[CPU_USER]
				+ stat->cpu_sysinfo.cpu[CPU_KERNEL]
				+ stat->cpu_sysinfo.cpu[CPU_IDLE]
				+ stat->cpu_sysinfo.cpu[CPU_WAIT]
				- cpu->user
				- cpu->kernel
				- cpu->idle
				- cpu->wait;
			double q = 1 / (double)(total) * 100;
			rates->user = (double)(stat->cpu_sysinfo.cpu[CPU_USER] - cpu->user) * q;
			UserP += rates->user / Ocpu;
			cpu->user = stat->cpu_sysinfo.cpu[CPU_USER];
			rates->kernel = (double)(stat->cpu_sysinfo.cpu[CPU_KERNEL] - cpu->kernel) * q;
			KernelP += rates->kernel / Ocpu;
			cpu->kernel = stat->cpu_sysinfo.cpu[CPU_KERNEL];
			rates->idle = (double)(stat->cpu_sysinfo.cpu[CPU_IDLE] - cpu->idle) * q;
			IdleP += rates->idle / Ocpu;
			cpu->idle = stat->cpu_sysinfo.cpu[CPU_IDLE];
			rates->wait = (double)(stat->cpu_sysinfo.cpu[CPU_WAIT] - cpu->wait) * q;
			IdleP += rates->wait / Ocpu;
			cpu->wait = stat->cpu_sysinfo.cpu[CPU_WAIT];
		}
	}
	if (num == Ocpu)
		return 0;
	assert(num < Ocpu);
	return 1;
}


static int
get_net_stats(void)
{
	class_chain_t *n = Net;
	NetInR = NetOutR = 0;
	while (n) {
		kstat_named_t *data;
		if (kstat_read(KS,n->kstat,0) == -1) {
			/* network interface has been taken offline */
			n->kstat = 0;
			return 1;
		}
		assert(n->kstat->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(n->kstat,"ipackets64");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (n->value[0])
			NetInR += (double)(data->value.ui64 - n->value[0]) / 
				(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9;
		n->value[0] = data->value.ui64;
		data = kstat_data_lookup(n->kstat,"opackets64");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (n->value[1])
			NetOutR += (double)(data->value.ui64 - n->value[1]) / 
				(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9;
		n->value[1] = data->value.ui64;
		n->timestamp = n->kstat->ks_snaptime;
		n = n->next;
	}
	return 0;
}


static void
setup_nets(void)
{
	class_chain_t *n, *p = 0;
	Net = get_class_chain("net",0,0,Net);
	n = Net;
	while (n) {
		/* The loopback interface "lo0" doesn't seem to have
		 * a ipackets64 field. So we use this field to be able
		 * to ignore it...
		if ((0 == strcmp(n->kstat->ks_name,"zero_copy")) ||
		    (0 == strcmp(n->kstat->ks_name,"lo0")) ||
		 */
		if ((-1 == kstat_read(KS,n->kstat,0)) ||
		    (0 == kstat_data_lookup(n->kstat,"ipackets64"))) {
			class_chain_t *l = n;
			if (p)
				p->next = n->next;
			else
				Net = n->next;
			n = n->next;
			free(l);
		} else
			n = n->next;
	}
}


static void
setup_cpus(void)
{
	unsigned num = 0, maxid = Maxid;
	int i;
	kstat_t *k;

#ifdef _SC_CPUID_MAX
	Maxid = sysconf(_SC_CPUID_MAX);
#else
	/* the maximum CPUID cannot be asked on Solaris 8 systems.
	 * Therefore, the value of NCPU in sys/param.h is used as 
	 * maximum CPUID, which should work for almost all cases.
	 * NCPU is a kernel limit for number of CPUs, _not_ maximum ID.
	 * So this is _not_ perfect, but good enough.
	 */
	Maxid = 56;
#endif
	
	/* number of cpu change during runtime */
	while (maxid > Maxid) {
		free(Cpus[maxid--]);
	}
	Cpus = realloc(Cpus,sizeof(cpudata_t *) * (Maxid + 1));
	assert(Cpus);
	while (maxid <= Maxid) {
		Cpus[maxid] = calloc(1,sizeof(cpudata_t));
		assert(Cpus[maxid]);
		++maxid;
	}
	assert(Ncpus != 0);
	for (i = 0; i <= Maxid; ++i) {
		Cpus[i]->cpustat = 0;
		Cpus[i]->cpusys = 0;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if (0 == strcmp(k->ks_module,"cpu_stat")) {
			assert(k->ks_type == KSTAT_TYPE_RAW);
			assert(k->ks_instance <= maxid);
			Cpus[k->ks_instance]->cpustat = k;
			if (Cpus[k->ks_instance]->rates == 0) {
				Cpus[k->ks_instance]->rates = calloc(1,sizeof(cpurates_t));
				assert(Cpus[k->ks_instance]->rates);
			}
			Cpus[k->ks_instance]->rates->id = k->ks_instance;
			dbug("found CPU with id %d\n",k->ks_instance);
			++num;
		} else if ((0 == strcmp(k->ks_module,"cpu"))
				&& (0 == strcmp(k->ks_name,"sys"))) {
			assert(k->ks_instance <= maxid);
			Cpus[k->ks_instance]->cpusys= k;
			assert(k->ks_type == KSTAT_TYPE_NAMED);
		}
	}
	assert(num <= Ncpus);
	if (Ocpu != num) {
		/* Try to reduce the transient effect of onlining of CPUs on
		 * metrics calculation.
		 */
		UserP /= Ocpu;
		UserP *= num;
		KernelP /= Ocpu;
		KernelP *= num;
		IdleP /= Ocpu;
		IdleP *= num;
		Ocpu = num;
		dbug("Ocpu = %d\n",Ocpu);
	}
}


static int
reconfigure(void)
{
	static kstat_t *sysmisc = 0;
	static kstat_named_t *data = 0;
	int upd;
	kid_t kid;
	class_chain_t *c;

	upd = kstat_chain_update(KS);
	if ((0 == upd) && Ncpus)
		return 1;
	if (-1 == upd)
		error("Kstat update failed: %s\n",strerror(errno));
	if (sysmisc == 0) {
		sysmisc = kstat_lookup(KS,"unix",0,"system_misc");
		assert(sysmisc && (sysmisc->ks_type == KSTAT_TYPE_NAMED));
	}
	kid = kstat_read(KS,sysmisc,0);
	assert(kid != -1);
	data = kstat_data_lookup(sysmisc,"ncpus");
	assert(data && data->data_type == KSTAT_DATA_UINT32);
	Ncpus = data->value.ui32;
	setup_cpus();
	Err = get_class_chain("device_error",0,0,Err);
	Disks = get_class_chain("disk","sd",0,Disks);
	assert(Disks);
	c = Disks;
	while (c->kstat && c->next)
		c = c->next;
	if (c->kstat) {
		if (c->next == 0)
			c->next = calloc(1,sizeof(class_chain_t));
		c = c->next;
	}
	assert(c->kstat == 0);
	(void) get_class_chain("disk","cmdk",0,c);
	Nfs = get_class_chain("nfs","nfs",0,Nfs);
	NfsServer = get_class_chain("misc","nfs","nfs_server",NfsServer);
	return 1;
}


int
update_fs_rates(void)
{
	static class_chain_t *fs_chain = 0;
	class_chain_t *c;
	fsstat_t *fs = get_fs_stats();
	fs_chain = get_class_chain(0,0,"vopstats_",fs_chain);
	c = fs_chain;
	while (c) {
		unsigned long nodeid;
		kstat_t *s = c->kstat;
		kstat_named_t *data;
		fsstat_t *f = fs;
		double td;
		unsigned long long ops = 0;

		if ((c->iteration == Iteration) || (s == 0)) {
			c = c->next;
			continue;
		}
		if (1 != sscanf(s->ks_name+9,"%lx",&nodeid)) {
			c = c->next;
			continue;
		}
		dbug("nodeid, %x\n",nodeid);
		while (f && (f->fsid != nodeid))
			f = f->next;
		if (f == 0) {
			c = c->next;
			continue;
		}
		dbug("vopstat: %s, nodeid, %x, fs: %s\n",s->ks_name,nodeid,f->dev);
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert((s->ks_type == KSTAT_TYPE_NAMED)
			&& (s->ks_snaptime != c->timestamp));

		data = kstat_data_lookup(s,"read_bytes");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		f->read_rate = (double)((data->value.ui64 - f->numread) >> 10) / 
				(double)(s->ks_snaptime - f->rate_ts) * 1E9;
		dbug("read: %llu-%llu/%llu-%llu, %g\n",data->value.ui64,f->numread,s->ks_snaptime,f->rate_ts,f->read_rate);
		f->numread = data->value.ui64;

		data = kstat_data_lookup(s,"readdir_bytes");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		f->read_rate += (double)((data->value.ui64 - f->numreaddir) >> 10) / 
				(double)(s->ks_snaptime - f->rate_ts) * 1E9;
		dbug("read: %llu-%llu/%llu-%llu, %g\n",data->value.ui64,f->numread,s->ks_snaptime,f->rate_ts,f->read_rate);
		f->numreaddir = data->value.ui64;

		data = kstat_data_lookup(s,"write_bytes");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		f->write_rate = (double)((data->value.ui64 - f->numwrite) >> 10) / 
				(double)(s->ks_snaptime - f->rate_ts) * 1E9;
		dbug("write: %llu-%llu, %g\n",data->value.ui64,f->numwrite,f->write_rate);
		f->numwrite = data->value.ui64;

		data = kstat_data_lookup(s,"naccess");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->naccess;
		dbug("naccess = %llu\n",data->value.ui64-f->naccess);
		f->naccess = data->value.ui64;

		data = kstat_data_lookup(s,"naddmap");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->naddmap;
		dbug("naddmap = %llu\n",data->value.ui64-f->naddmap);
		f->naddmap = data->value.ui64;

		data = kstat_data_lookup(s,"nclose");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nclose;
		dbug("nclose = %llu\n",data->value.ui64-f->nclose);
		f->nclose = data->value.ui64;

		data = kstat_data_lookup(s,"ncmp");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ncmp;
		dbug("ncmp = %llu\n",data->value.ui64-f->ncmp);
		f->ncmp = data->value.ui64;

		data = kstat_data_lookup(s,"ncreate");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ncreate;
		dbug("ncreate = %llu\n",data->value.ui64-f->ncreate);
		f->ncreate = data->value.ui64;

		data = kstat_data_lookup(s,"ndelmap");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndelmap;
		dbug("ndelmap = %llu\n",data->value.ui64-f->ndelmap);
		f->ndelmap = data->value.ui64;

		data = kstat_data_lookup(s,"ndispose");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndispose;
		dbug("ndispose = %llu\n",data->value.ui64-f->ndispose);
		f->ndispose = data->value.ui64;
		
		data = kstat_data_lookup(s,"ndump");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndump;
		dbug("ndump = %llu\n",data->value.ui64-f->ndump);
		f->ndump = data->value.ui64;

		data = kstat_data_lookup(s,"ndumpctl");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndumpctl;
		dbug("ndumpctl = %llu\n",data->value.ui64-f->ndumpctl);
		f->ndumpctl = data->value.ui64;

		data = kstat_data_lookup(s,"nfid");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nfid;
		dbug("nfid = %llu\n",data->value.ui64-f->nfid);
		f->nfid = data->value.ui64;

		data = kstat_data_lookup(s,"nfrlock");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nfrlock;
		dbug("nfrlock = %llu\n",data->value.ui64-f->nfrlock);
		f->nfrlock = data->value.ui64;

		data = kstat_data_lookup(s,"nfsync");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nfsync;
		dbug("nfsync = %llu\n",data->value.ui64-f->nfsync);
		f->nfsync = data->value.ui64;

		data = kstat_data_lookup(s,"ngetattr");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ngetattr;
		dbug("ngetattr = %llu\n",data->value.ui64-f->ngetattr);
		f->ngetattr = data->value.ui64;

		data = kstat_data_lookup(s,"ngetpage");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ngetpage;
		dbug("ngetpage = %llu\n",data->value.ui64-f->ngetpage);
		f->ngetpage = data->value.ui64;

		data = kstat_data_lookup(s,"ngetsecattr");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ngetsecattr;
		dbug("ngetsecattr = %llu\n",data->value.ui64-f->ngetsecattr);
		f->ngetsecattr = data->value.ui64;

		data = kstat_data_lookup(s,"ninactive");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ninactive;
		dbug("ninactive = %llu\n",data->value.ui64-f->ninactive);
		f->ninactive = data->value.ui64;

		data = kstat_data_lookup(s,"nioctl");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nioctl;
		dbug("nioctl = %llu\n",data->value.ui64-f->nioctl);
		f->nioctl = data->value.ui64;

		data = kstat_data_lookup(s,"nlink");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nlink;
		dbug("nlink = %llu\n",data->value.ui64-f->nlink);
		f->nlink = data->value.ui64;

		data = kstat_data_lookup(s,"nlookup");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nlookup;
		dbug("nlookup = %llu\n",data->value.ui64-f->nlookup);
		f->nlookup = data->value.ui64;

		data = kstat_data_lookup(s,"nmap");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nmap;
		dbug("nmap = %llu\n",data->value.ui64-f->nmap);
		f->nmap = data->value.ui64;

		data = kstat_data_lookup(s,"nmkdir");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nmkdir;
		dbug("nmkdir = %llu\n",data->value.ui64-f->nmkdir);
		f->nmkdir = data->value.ui64;

		data = kstat_data_lookup(s,"nopen");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nopen;
		dbug("nopen = %llu\n",data->value.ui64-f->nopen);
		f->nopen = data->value.ui64;

		data = kstat_data_lookup(s,"npageio");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->npageio;
		dbug("npageio = %llu\n",data->value.ui64-f->npageio);
		f->npageio = data->value.ui64;

		data = kstat_data_lookup(s,"npathconf");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->npathconf;
		dbug("npathconf = %llu\n",data->value.ui64-f->npathconf);
		f->npathconf = data->value.ui64;

		data = kstat_data_lookup(s,"npoll");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->npoll;
		dbug("npoll = %llu\n",data->value.ui64-f->npoll);
		f->npoll = data->value.ui64;

		data = kstat_data_lookup(s,"nputpage");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nputpage;
		dbug("nputpage = %llu\n",data->value.ui64-f->nputpage);
		f->nputpage = data->value.ui64;

		data = kstat_data_lookup(s,"nread");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nread;
		dbug("nread = %llu\n",data->value.ui64-f->nread);
		f->nread = data->value.ui64;

		data = kstat_data_lookup(s,"nreaddir");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nreaddir;
		dbug("nreaddir = %llu\n",data->value.ui64-f->nreaddir);
		f->nreaddir = data->value.ui64;

		data = kstat_data_lookup(s,"nreadlink");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nreadlink;
		dbug("nreadlink = %llu\n",data->value.ui64-f->nreadlink);
		f->nreadlink = data->value.ui64;

		data = kstat_data_lookup(s,"nrealvp");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nrealvp;
		dbug("nrealvp = %llu\n",data->value.ui64-f->nrealvp);
		f->nrealvp = data->value.ui64;

		data = kstat_data_lookup(s,"nremove");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nremove;
		dbug("nremove = %llu\n",data->value.ui64-f->nremove);
		f->nremove = data->value.ui64;

		data = kstat_data_lookup(s,"nrename");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nrename;
		dbug("nrename = %llu\n",data->value.ui64-f->nrename);
		f->nrename = data->value.ui64;

		data = kstat_data_lookup(s,"nrmdir");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nrmdir;
		dbug("nrmdir = %llu\n",data->value.ui64-f->nrmdir);
		f->nrmdir = data->value.ui64;

		data = kstat_data_lookup(s,"nrwlock");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nrwlock;
		dbug("nrwlock = %llu\n",data->value.ui64-f->nrwlock);
		f->nrwlock = data->value.ui64;

		data = kstat_data_lookup(s,"nrwunlock");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nrwunlock;
		dbug("nrwunlock = %llu\n",data->value.ui64-f->nrwunlock);
		f->nrwunlock = data->value.ui64;

		data = kstat_data_lookup(s,"nseek");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nseek;
		dbug("nseek = %llu\n",data->value.ui64-f->nseek);
		f->nseek = data->value.ui64;

		data = kstat_data_lookup(s,"nsetattr");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nsetattr;
		dbug("nsetattr = %llu\n",data->value.ui64-f->nsetattr);
		f->nsetattr = data->value.ui64;

		data = kstat_data_lookup(s,"nsetfl");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nsetfl;
		dbug("nsetfl = %llu\n",data->value.ui64-f->nsetfl);
		f->nsetfl = data->value.ui64;

		data = kstat_data_lookup(s,"nsetsecattr");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nsetsecattr;
		dbug("nsetsecattr = %llu\n",data->value.ui64-f->nsetsecattr);
		f->nsetsecattr = data->value.ui64;

		data = kstat_data_lookup(s,"nshrlock");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nshrlock;
		dbug("nshrlock = %llu\n",data->value.ui64-f->nsetsecattr);
		f->nshrlock = data->value.ui64;

		data = kstat_data_lookup(s,"nspace");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nspace;
		dbug("nspace = %llu\n",data->value.ui64-f->nsetsecattr);
		f->nspace = data->value.ui64;

		data = kstat_data_lookup(s,"nsymlink");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nsymlink;
		dbug("nsymlink = %llu\n",data->value.ui64-f->nsymlink);
		f->nsymlink = data->value.ui64;

		data = kstat_data_lookup(s,"nvnevent");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nvnevent;
		dbug("nvnevent = %llu\n",data->value.ui64-f->nvnevent);
		f->nvnevent = data->value.ui64;

		data = kstat_data_lookup(s,"nwrite");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nwrite;
		dbug("nwrite = %llu\n",data->value.ui64-f->nwrite);
		f->nwrite = data->value.ui64;

		f->ops_rate = (double)(ops) / 
				(double)(s->ks_snaptime - f->rate_ts) * 1E9;
		dbug("%s: oprate %g, ntotal %llu, ops %llu, dt %lld\n",f->dev,f->ops_rate,f->ntotal,ops,s->ks_snaptime-f->rate_ts);
		f->ntotal = ops;

		f->rate_ts = s->ks_snaptime;
		c->iteration = Iteration;
		dbug("%s: I/O(%g/%g), %lld %lld\n",f->dev,f->read_rate,f->write_rate,c->value[0],c->value[1]);
		c = c->next;
	}
	while (fs) {
		if (fs->zroot) {
			fsstat_t *z = fs->zroot;
			dbug("accounting %s in %s\n",fs->dev,z->dev);
			z->write_rate += fs->write_rate; 
			z->read_rate += fs->read_rate; 
			z->ops_rate += fs->ops_rate; 
		}
		fs = fs->next;
	}
	return 0;
}


void
get_local_data(sysstat_t *s)
{
	IntrR = IntrBlkR = IntrThreadR = XcallsR = 0;
	LockfR = 0;
	SyscallR = SysexecR = SysforkR = 0;
	SysreadR = SyswriteR = 0;
	InvSwtchR = PswitchR = MigrateR = 0;
	IdleP = UserP = KernelP = 0;
	DiskIR = DiskOR = 0;
	NfsdR = NfsIR = NfsOR = 0;
	ScanR = PgswapinR = PgswapoutR = 0;
	AsFaultR = HatFaultR = MajFaultR = 0;
	AnonfreeR = AnonpginR = AnonpgoutR = 0;
	Cpurates = 0;
	while (reconfigure()) {
		clock_t now;

		if (get_cpu_stat())
			continue;
		if (get_disk_io())
			continue;
		if (update_fs_stats())
			continue;
		if (update_fs_rates())
			continue;
		if (get_nfs_io())
			continue;
		if (get_unix_systempages())
			continue;
		if (Ipstat) {
			/* Solaris 10 extras */
			if (get_cpu_sys())
				continue;
			get_ip_ip();
		} else if (get_net_stats())
			continue;
		get_device_errors();
		get_swapinfo();
		++Iteration;
		if (s == 0)
			return;
		now = time(0);
		s->TS = now;
		s->BootTime = BootTime;
		getloadavg(s->loadavg,3);
		s->Ncpus = Ncpus;
		s->Ocpus = Ocpu;
		s->Devices = Devices;
		s->Storage = Storage;
		s->Zsize = Zsize;
		s->Zfree = Zfree;
		s->Zpools = Zpools;
		s->HardErrors = HardErrors;
		s->SoftErrors = SoftErrors;
		s->Physmem = Physmem;
		s->Pagesfree = Pagesfree;
		s->Pagestotal = Pagestotal;
		s->Pageslocked = Pageslocked;
		s->SwapTotal = SwapTotal;
		s->SwapFree = SwapFree;
		s->ScanR = ScanR;
		s->AnonpginR = AnonpginR;
		s->AnonpgoutR = AnonpgoutR;
		s->AnonfreeR = AnonfreeR;
		s->AsFaultR = AsFaultR;
		s->HatFaultR = HatFaultR;
		s->MajFaultR = MajFaultR;
		s->PgswapinR = PgswapinR;
		s->PgswapoutR = PgswapoutR;
		s->IntrR = IntrR;
		s->IntrBlkR = IntrBlkR;
		s->IntrThreadR = IntrThreadR;
		s->XcallsR = XcallsR;
		s->LockfR = LockfR;
		s->SyscallR = SyscallR;
		s->SysexecR = SysexecR;
		s->SysforkR = SysforkR;
		s->SysreadR = SysreadR;
		s->SyswriteR = SyswriteR;
		s->InvSwtchR = InvSwtchR;
		s->PswitchR = PswitchR;
		s->MigrateR = MigrateR;
		s->IdleP = IdleP;
		s->UserP = UserP;
		s->KernelP = KernelP;
		s->DiskIR = DiskIR/1024;
		s->DiskOR = DiskOR/1024;
		s->NfsIR = NfsIR/1024;
		s->NfsOR = NfsOR/1024;
		s->NfsdR = NfsdR;
		s->FsFree = FsFree;
		s->IpInR = IpInR;
		s->IpOutR = IpOutR;
		s->NetInR = NetInR;
		s->NetOutR = NetOutR;
		s->IpInDiscards = IpInDiscards;
		s->IpOutDiscards = IpOutDiscards;
		s->IpInErrs = IpInErrs;
		s->fs = get_fs_stats();
		s->us = get_procinfo();
		s->cpurates = Cpurates;
		s->displayed = 0;
		return;
	}
}


void
init_localstat(sysstat_t *s)
{
	int ret;
	char *next;
	struct rlimit rlimit;

	if (s == 0)
		s = &Sysstat;
	if (-1 == getrlimit(RLIMIT_NOFILE,&rlimit))
		info("could not get resource limit: %s\n",strerror(errno));
	rlimit.rlim_cur = rlimit.rlim_max;
	if (-1 == setrlimit(RLIMIT_NOFILE,&rlimit))
		warn("could not reset no of files limit: %s\n",strerror(errno));
	get_boottime();
	KS = kstat_open();
	if (KS == 0)
		error("cannot access kernel statistics: %s\n",strerror(errno));
	Syspages = kstat_lookup(KS,"unix",0,"system_pages");
	assert(Syspages && (Syspages->ks_type == KSTAT_TYPE_NAMED));
	Ipstat = kstat_lookup(KS,"ip",0,"ip");
	if (Ipstat == 0) {
		/* before Solaris 10 */
		IpInR = IpOutR = -1;
		setup_nets();
	} else 
		assert(Ipstat->ks_type == KSTAT_TYPE_NAMED);
	Pgsz = getpagesize();
	ret = gethostname(s->Hostinfo,sizeof(s->Hostinfo));
	assert(ret == 0);
	if ((next = strchr(s->Hostinfo,'.')) == 0)
		next = s->Hostinfo + strlen(s->Hostinfo);
	*next = ' ';
	++next;
	sysinfo(SI_SYSNAME,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	next += strlen(next);
	*next = ' ';
	++next;
	sysinfo(SI_RELEASE,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	next += strlen(next);
	*next = ' ';
	++next;
	sysinfo(SI_PLATFORM,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	reconfigure();
	/* getdata once then sleep 0.3sec to get reasonable start values */
	get_local_data(0);
	(void) usleep(300000);	
}


void
daemon_sighandler(int s)
{
	assert(s == SIGTERM || s == SIGQUIT || s == SIGINT || s == SIGPIPE);
	if (Localmode == 0)
		(void) unlink("/tmp/sysstat.pid");
	exit(EXIT_SUCCESS);
}


void
rundaemon(int interval,int count)
{
	int c = 0;
	(void) signal(SIGTERM,daemon_sighandler);
	(void) signal(SIGQUIT,daemon_sighandler);
	(void) signal(SIGINT,daemon_sighandler);
	if (Localmode)
		(void) signal(SIGPIPE,daemon_sighandler);
	init_zfs();
	init_localstat(&Sysstat);
	for (;(count == 0) || (c != count); count ? ++c : c) {
		get_local_data(&Sysstat);
		send_to_clients(&Sysstat,Localmode);
		(void) sleep(interval);
	}
}

// vim:tw=0
